//
// NOTE: If you're using Xcode and seeing raw markdown contents, go to the
// Editor menu -> Show Rendered Markup.
//

/*:
# Swift + Python Interoperability Tutorial

 This tutorial shows how to use the PythonGlue.swift code to talk to arbitrary
 Python APIs from Swift.  The implementation approach is completely dynamic -
 it doesn't require invasive Swift compiler support nor does it require any
 changes to Python.

 In the examples below, I show how to do something in Swift, with the
 corresponding Python code in a comment right above it.

 The first thing to know is that all of the Python builtin primitives
 (`print`, `eval`, `vars`, etc) are exposed on a top level `Python` type - we
 scope them so they don't affect unsuspecting Swift code.  To get started, you'd
 typically import a Python module:
*/

// Python: import numpy as np
let np = Python.import("numpy")

/*:
 In Python, "everything is an object", including modules.  In this example,
 the type of the returned `np` value is `PyVal`, which is the general Python
 object representation in Swift.  Python objects allow member access, calling,
 subscripting, and a variety of other behavior, which we reflect into Swift
 syntax.  For example, if you wanted to print out the module above using the
 Python version of print, you'd call the print builtin, and pass it as an
 argument:
*/

// Python: print(np)
Python.print.call(args: np)      // -> Prints in playground console.

// Of course you can also use the Swift print function!
print("Swift can print too: \(np)")

/*:
 In this example, `Python.print` is also a value of the dynamic `PyVal` value,
 and we're using the call method to invoke it, passing `np`.  We expect this to
 be sugared into a direct Swift call at some point, which will allow us to
 write:

     Python.print(np)

 That said, we don't have that yet, so we'll continue without it - using the
more verbose and irritating syntax that does work today.

 ## Numpy Tutorial Examples

 To give a couple more examples of how to use this, we'll follow the numpy
 tutorial from https://docs.scipy.org/doc/numpy-dev/user/quickstart.html

 Python "declarations" are actually just properties in the module they are
 defined in so we can start directly accessing them:
*/

// Python: np.arange(15)
np.get(member: "arange").call(args: 15)

/*:
 This is the very direct code that accesses the `arange` property on the
 numpy module.  arange happens to be a function, which we then call, passing
 in a list of Python values as arguments.  You can even see the result of this
 call in the Playground result sidebar to the right, since `PyVal`'s are wired
 into Playgrounds and the Swift REPL.

 At this point it is worth observing that many Swift values (like 15 above) can
 be directly passed to Python.  This is because parameters, subscript indices,
 and other places take anything that conforms to the `PythonConvertible` type
 (which the builtin types conform to).  There are also failable initializers
 that convert back the other ways through normal `Int(somePyVal)` syntax.

 All that said, the syntax above is obviously overwhelming to do something
 things like this.  As with calls, we expect that Swift will eventually add a
 tiny bit of syntactic sugar to make the member access nicer, which will allow
 us to write:

     let a = np.get(member: "arange").call(args: 15) // Today
     let a = np.arange.call(args: 15)  // DynamicMemberLookup proposal
     let a = np.arange(15)             // DynamicMemberLookup+DynamicCallable

 we don't have these features yet, so we added a call+memberlookup method to
 handle the common case above.  It is important to know that this is simply a
 composition that wraps the two operations, one which we expect to go away
 completely once the sugar is available.  With it, we can write somewhat more
 complex examples, such as:
 */

// Python:        a = np.arange(15).reshape(3, 5)
// Future Swift:  let a = np.arange(15).reshape(3, 5)
let a = np.call(member: "arange", args: 15).call(member: "reshape", args: 3, 5)


/*:
 We can continue following the tutorial above:
 */

// Python:        a.shape
// Future Swift:  a.shape
a.get(member: "shape")

// Python:        a.ndim
// Future Swift:  a.ndim
a.get(member: "ndim")

// Python:        a.dtype.name
// Future Swift:  a.dtype.name
a.get(member: "dtype").get(member: "name")

// Python:        a.itemsize
// Future Swift:  a.itemsize
a.get(member: "itemsize")

// Python:        a.size
// Future Swift:  a.size
a.get(member: "size")

// Python:        type(a)
// Future Swift:  Python.type(a)
Python.type.call(args: a)

/*:
 We can even pass aggregate types like arrays into Python, but until we have
 conditional conformances, we have to wrap them with a call to build them,
 named `PyVal(array:...)` (yes, Python programmers call them lists.  The target
 audience is Swift programmers though, and they call them arrays):
 */

// Python:        b = np.array([6, 7, 8])
// Future Swift:  let b = np.array([6, 7, 8])
let b = np.call(member: "array", args: PyVal(array: 6, 7, 8))

// Python:        type(b)
// Future Swift:  Python.type(b)
Python.type.call(args: b)

/*:
 ## Python Exceptions in Swift

 We could continue through the NumPy tutorial, but hopefully it is clear that
 APIs "just work" if they are built on member access and calls (which is pretty
 much everything).  What are the exceptions?

 Well since you coincidentally ask, we've glossed over the fact that almost all
 calls in Python can throw an exception, and while it is usually ignored
 by most code, it is occasionally useful to be able to catch and process them.
 To do this, we have a `.throwing` adaptor that you can use to tell the system
 that you want to handle an error if one is produced.  For example, if you want
 to test to see if a file can be opened, you would do something like this:
 */

// Python code:
// try:
//   x = open("/file/that/doesnt/exist")
//   print(x)
// except:
//   print("file not found, just like we expected!")
//
do {
  // Future Swift: let x = try Python.open.throwing("/file/that/doesnt/exist")
  let x = try Python.open.throwing.call(args: "/file/that/doesnt/exist")
  print(x)
} catch let err {
  print("file not found, just like we expected!")

  // Here is the error we got:
  print(err)
}

/*:
 Though this might seem weird on first glance, it actually works out really well
 in practice, because the `try` markers in Swift are *required* when something
 throws, and *produce a warning* when they are inappropriately placed.  Further,
 the API on the `ThrowingPyVal` type only supports the few operations that can
 throw, so you can't just put this anywhere: if you put it before the member
 instead of before the call, you'd get a compile time error.  This combination
 of behavior makes the intent of the Python code really clear and catches common
 programmer mistakes immediately.

 Property and subscript lookups can also fail by throwing an exception in
 Python.  This isn't he tradition in Swift though: optional chaining makes it
 far more natural to handle failable property accesses by having them return
 optionals.  Also, Swift properties and subscripts can't throw Swift errors.

 As such, we represent failable member lookups with a `checking` adaptor, which
 works the same was as `throwing` does for calls:
*/

// Future Swift: let c = b.checking.missingMethod?(42)
let c = b.checking.get(member: "missingMethod")?.call(args: 42)
assert(c == nil)

/*:
 It is worth mentioning that the syntax above could be sugared even further by
 introducing two postfix operators to represent these operations, allowing
 syntax like this:

     let x = try Python.open^("/file/that/doesnt/exist")

 While it is possible to do this, I think it is a bad idea to do so for several
 reasons (most significant to least):

   1) If we add this early, it will be difficult to take away in the future. It
      is better to start small.
   2) Exceptions are really rare in Python programs, and ones that are intended
      to be recovered from are even more so.  We shouldn't be too worried about
      sugaring them: having them stand out is actually a good thing.
   3) It looks weird, is hard to read, and completely impenetrable for anyone
      who hasn't memorized this.

 ## Keyword Arguments in Calls

 Like Swift, Python supports keyword arguments in function calls.  We support
 this by allowing you to provide an array of `(String, PyVal)` tuples at call
 sites.  Future language support can sugar this into Swift keyword arguments
 directly:
 */

// Python:        d = np.array([1, 2, 3], dtype="i2")
// Future Swift:  let d = np.array([6, 7, 8], dtype: "i2")
let d = np.call(member: "array", args: PyVal(array: 6, 7, 8),
                kwargs: [("dtype", "i2")])

/*:
 ## Python Lists, Dictionaries, Slices and Tuples oh my!

 We mentioned above that you can pass most primitive Swift types into Python
 directly (because they conform to the `PythonConvertible` protocol, but right
 now most aggregate types (arrays, dictionaries, ranges, and tuples) cannot.
 Fortunately, arrays, dictionaries, and ranges cannot be passed because of a
 soon-to-be-resolved limitation in the Swift generics system: the problem is
 that types like Array cannot conditionally conform to the PythonConvertible protocol when
 their elements conform.  Major progress is underway on Swift master, so we
 expect this to be resolved soon.

 In the mean time, we provide helper functions to produce these values:
 `PyVal(array:)` and `PyVal(dict:)`, each of which can take a list of values or a
 collection containing them:
 */

// Python:        array = [1.0, 2.0, 3.0, 4.0]
// Future Swift:  let array = [1.0, 2.0, 3.0, 4.0] as PyVal
let array = PyVal(array: 1.0, 2.0, 3.0, 4.0)

// Python:        dict = {"x": 1.0, "y": 2.0}
// Future Swift:  let dict = ["x": 1.0, "y": 2.0] as PyVal
let dict = PyVal(dict: ("x", 1.0), ("y", 2.0))

// Python collections can be heterogenous too of course, and may be initialized
// from existing Swift collections:
let array2 = PyVal(arrayContentsOf: [a, 1, "foo"])
let dict2 = PyVal(dictContentsOf: ["x": 42, "y" : "something else"])

/*:
 We handle Python slices the same way right now, but we should definitely make
 Swift ranges convert to them.  Python slices support striding and negative
 indexing, so there are some design questions to answer.  Right now, you can
 just use the `PyVal(slice: start, end, stride)` helper:
 */

// Python:        slice(1, 2, 0.5)
// Python:        x[1:2:0.5]          // Context sensitive Python sugar.
// Future Swift:  (1..<2).by(0.5)     // Hypothetical example.
let slice = PyVal(slice: 1, 2, 0.5)

/*:
 Tuples are harder for a couple of reasons: they cannot conform to protocols
 and they are variadic.  There is a theoretical future where Swift fixes both
 of these problems, but that isn't a short-term future.  Fortunately, tuples are
 rare enough in Python code that they aren't worth worrying too much about.

 While we don't have to worry about them, we do need to provide basic support
 for them for the APIs that do use them.  To support for both creating tuples
 and extracting elements out of them, we provide the following facilities:
 */

// Python: tuple = (1, "foo", a)
let tuple = PyVal(tuple: 1, "foo", a)

// Python: print(tuple[1])
print(tuple[1])

// Python: (t1, t2, t3) = tuple
let (t1, t2, t3) = tuple.get3Tuple()
t1
t2
t3

/*:
 The later example shows how we handle destructuring of tuples, which is a
 bit ugly, but works fine for the rare cases it is needed.  Better naming for
 these methods would be great though.

 ## Going Further

 There are still pieces that are missing from this interop layer that I haven't
 gotten around to adding yet, and more design topics to consider.  Things that
 come to mind:

   1) Overloading the Python operator set to work on PyVal's.
   2) Making PyVal be a Swift Sequence so you can enumerate and for/in them.
   3) Making PyVal conform to the literal types, so you
      can initialize them directly from integer, floating
      point, array, and other literal types.
   4) Consider modeling failable subscript and property accesses as returning
      IUOs.  This could simplify the model a lot by eliminating the
      'checking' modifier, at the cost of less explicitness and all the danger
      of IUOs.
 */
